/* slider.h */
/* Pavel Surynek */

#ifndef __SLIDER_H__
#define __SLIDER_H__

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <time.h>
#include <malloc.h>
#include <limits.h>
#include <math.h>

#include <map>
#include <set>

#define MAX_DEPTHS_COUNT    2048
#define SIGNS_COUNT          256
#define BI_SIGNS_COUNT       256
#define TRI_SIGNS_COUNT      256


extern long STEPS;
extern int SOLUTION_DEPTH;
extern int MAX_DEPTHS[MAX_DEPTHS_COUNT];


typedef struct
{
	int m_V;
	int *m_vrtxs;
	int *m_goal;

	int m_C1;
	int *m_cycle1;

	int m_C2;
	int *m_cycle2;

	int m_C3;
	int *m_cycle3;

	int m_H1;
	int *m_hash1;

	int m_H2;
	int *m_hash2;

	int m_H3;
	int *m_hash3;

	int m_X;
	int m_Y;

	int m_A, m_B, m_C;
} SLIDER;


typedef struct
{
	int m_V;
	int **m_dist;
} DIST_MATRIX;


typedef enum 
{
	MOVE_UNDEFINED,
	CYCLE_1_LEFT,
	CYCLE_1_RIGHT,
	CYCLE_2_LEFT,
	CYCLE_2_RIGHT,
	CYCLE_3_LEFT,
	CYCLE_3_RIGHT
} MOVE;


typedef struct
{
	int m_M, m_maxM;
	MOVE *m_moves;
} SOLUTION;


typedef struct
{
	int m_c1_sign;
	int m_c2_sign;
} BI_SIGNATURE;


typedef struct
{
	int m_c1_sign;
	int m_c2_sign;
	int m_c3_sign;
} TRI_SIGNATURE;


#define UNUSED(X)


class QUAD_SIGNATURE
{
public:
	bool operator==(const QUAD_SIGNATURE &q_sig) const
	{
		return (   m_c1_sign == q_sig.m_c1_sign
			&& m_c2_sign == q_sig.m_c2_sign
			&& m_c3_sign == q_sig.m_c3_sign
			&& m_c4_sign == q_sig.m_c4_sign);
	}

	bool operator<(const QUAD_SIGNATURE &q_sig) const
	{
		return ((m_c1_sign < q_sig.m_c1_sign) ? true : ((m_c1_sign > q_sig.m_c1_sign) ? false :
			((m_c2_sign < q_sig.m_c2_sign) ? true : ((m_c2_sign > q_sig.m_c2_sign) ? false :
			 ((m_c3_sign < q_sig.m_c3_sign) ? true : ((m_c3_sign > q_sig.m_c3_sign) ? false :
			  ((m_c4_sign < q_sig.m_c4_sign) ? true : ((m_c4_sign > q_sig.m_c4_sign) ? false : false))))))));
	}

public:
	int m_c1_sign;
	int m_c2_sign;
	int m_c3_sign;
	int m_c4_sign;
};


int calc_signature(int *init, int *goal, int V);
void calc_bi_signature(SLIDER *init, SLIDER *goal, BI_SIGNATURE *bi_sign);
void calc_tri_signature(SLIDER *init, SLIDER *goal, TRI_SIGNATURE *tri_sign);
void calc_quad_signature(SLIDER *init, SLIDER *goal, QUAD_SIGNATURE *quad_sign);

void print_signatures(int V);
void print_bi_signatures(int V);
double analyze_tri_signatures(int V);

void mult_dist_matrix(DIST_MATRIX *dm);
void print_dist_matrix(DIST_MATRIX *dm);
void init_dist_matrix(SLIDER *sl, DIST_MATRIX *dm);

void init_slider(int UNUSED(c1_a), int c1_b, int c2_a, int c2_b, SLIDER *slider);
void destroy_slider(SLIDER *slider);
void copy_slider(const SLIDER *slider, SLIDER *copy);
void swap_slider(SLIDER *slider);
void swap_goal(int s1, int s2, SLIDER *slider);
int check_goal(SLIDER *slider);
int estimate_depth(SLIDER *slider, DIST_MATRIX *dist_matrix);
void print_slider(SLIDER *slider);
void fprint_slider(SLIDER *slider, FILE *f);

void cycle_1_rotate_L(SLIDER *slider);
void cycle_1_rotate_R(SLIDER *slider);
void cycle_2_rotate_L(SLIDER *slider);
void cycle_2_rotate_R(SLIDER *slider);
void cycle_3_rotate_L(SLIDER *slider);
void cycle_3_rotate_R(SLIDER *slider);

void init_solution(int max_moves, SOLUTION *solution);
void reset_solution(SOLUTION *solution);
void destroy_solution(SOLUTION *solution);

void push_move(MOVE move, SOLUTION *solution);
void pop_move(SOLUTION *solution);
void print_solution(SOLUTION *solution);
void fprint_solution(SOLUTION *solution, FILE *f);
int check_symmetry(SOLUTION *solution);
int solve_slider(int depth, int max_depth, SLIDER *slider, DIST_MATRIX *dist_matrix, SOLUTION *solution);
int solve_slider_R(int depth, int max_depth, SLIDER *slider, DIST_MATRIX *dist_matrix, SOLUTION *solution);
void reset_estimations(void);

int solve_slider_deepening(int id, double time_out, SLIDER *slider1, SLIDER *slider2, SOLUTION *solution1, SOLUTION *UNUSED(solution2), FILE *UNUSED(f));
void generate_infinity(SLIDER *slider1, SLIDER *slider2, SOLUTION *solution1, SOLUTION *solution2);
bool solve_slider_ID_PL(SLIDER *slider, SOLUTION *solution, double time_out = 10.0);

class STAR_NODE
{
public:
	STAR_NODE(const SLIDER *slider, int estimated, int depth)
	{
		copy_slider(slider, &m_slider);
		m_estimated = estimated;
		m_depth = depth;
	}

	STAR_NODE(const STAR_NODE &star_node)
	{
		copy_slider(&star_node.m_slider, &m_slider);
		m_estimated = star_node.m_estimated;
		m_depth = star_node.m_depth;
	}

	const STAR_NODE& operator=(const STAR_NODE &star_node)
	{
		destroy_slider(&m_slider);
		copy_slider(&star_node.m_slider, &m_slider);
		m_estimated = star_node.m_estimated;
		m_depth = star_node.m_depth;

		return *this;
	}

	virtual ~STAR_NODE()
	{
		destroy_slider(&m_slider);
	}

	bool operator==(const STAR_NODE &star) const
	{
		if (m_slider.m_V != star.m_slider.m_V)
		{
			return false;
		}
		for (int i = 0; i < m_slider.m_V; ++i)
		{
			if (m_slider.m_vrtxs[i] != star.m_slider.m_vrtxs[i])
			{
				return false;
			}
		}
		return true;
	}

	bool operator<(const STAR_NODE &star) const
	{
		return ((m_estimated < star.m_estimated) ? true : false);
	}

public:
	int m_estimated;
        int m_depth;
	SLIDER m_slider;
};

typedef std::map<QUAD_SIGNATURE, int, std::less<QUAD_SIGNATURE> > QUAD_SIGN_DEPTHS_map;
typedef std::multiset<STAR_NODE, std::less<STAR_NODE> > STAR_LIST_set;

bool is_closed(const STAR_NODE &star_node);
bool solve_star_loop(DIST_MATRIX *dist_matrix, clock_t begin = 0, double timeout = 10.0);
bool solve_star(SLIDER *slider, clock_t begin = 0, double timeout = 10.0);
void generate_star_infinity(SLIDER *slider1, SLIDER *slider2, SOLUTION *solution1, SOLUTION *solution2);

bool solve_slider_IDA(int depth, int max_estimation, int &new_estimation, SLIDER *slider, DIST_MATRIX *dist_matrix, SOLUTION *solution, clock_t begin = 0, double timeout = 10.0);
bool solve_IDA(SLIDER *slider, SOLUTION *solution, clock_t begin = 0, double timeout = 10.0);
void generate_IDA_infinity(SLIDER *slider1, SLIDER *slider2, SOLUTION *solution1, SOLUTION *solution2);

bool solve_slider_IDA_PL(int depth, int max_estimation, int &new_estimation, SLIDER *slider, DIST_MATRIX *dist_matrix, SOLUTION *solution, clock_t begin = 0, double timeout = 10.0);
bool solve_IDA_PL(SLIDER *slider, SOLUTION *solution, clock_t begin = 0, double timeout = 10.0);
void generate_IDA_PL_infinity(SLIDER *slider1, SLIDER *slider2, SOLUTION *solution1, SOLUTION *solution2);

#endif
